package ru.dbotthepony.kstarbound.world.entities.tile

import ru.dbotthepony.kommons.util.AABB
import ru.dbotthepony.kommons.vector.Vector2d
import ru.dbotthepony.kommons.vector.Vector2i
import ru.dbotthepony.kstarbound.defs.tile.TileDamage
import ru.dbotthepony.kstarbound.network.syncher.networkedSignedInt
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.entities.AbstractEntity

/**
 * (Hopefully) Static world entities (Plants, Objects, etc), which reside on cell grid
 */
abstract class TileEntity(path: String) : AbstractEntity(path) {
	protected val xTilePositionNet = networkedSignedInt()
	protected val yTilePositionNet = networkedSignedInt()

	init {
		xTilePositionNet.addListener(::updateSpatialIndex)
		yTilePositionNet.addListener(::updateSpatialIndex)
	}

	abstract val metaBoundingBox: AABB

	private fun updateSpatialIndex() {
		val spatialEntry = spatialEntry ?: return
		spatialEntry.fixture.move(metaBoundingBox + position)
	}

	var xTilePosition: Int
		get() = xTilePositionNet.get()
		set(value) {
			if (isSpawned) {
				xTilePositionNet.accept(world.geometry.x.cell(value))
			} else {
				xTilePositionNet.accept(value)
			}
		}

	var yTilePosition: Int
		get() = yTilePositionNet.get()
		set(value) {
			if (isSpawned) {
				yTilePositionNet.accept(world.geometry.x.cell(value))
			} else {
				yTilePositionNet.accept(value)
			}
		}

	var tilePosition: Vector2i
		get() = Vector2i(xTilePosition, yTilePosition)
		set(value) {
			xTilePosition = value.x
			yTilePosition = value.y
		}

	override val position: Vector2d
		get() = Vector2d(xTilePosition.toDouble(), yTilePosition.toDouble())

	abstract val occupySpaces: Set<Vector2i>
	abstract fun damage(damageSpaces: List<Vector2i>, source: Vector2d, damage: TileDamage): Boolean

	override fun onJoinWorld(world: World<*, *>) {
		updateSpatialIndex()
	}

	override fun onRemove(world: World<*, *>, isDeath: Boolean) {
	}
}
